// Copyright © 2015, 2016 T-semi
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

//コメントアウト解除で標準出力に実行時間に関する情報を出力。OFF推奨
//#define FINDER_GET_EXECUTION_TIME

#include "finder.h"

#include <stdint.h>

#include <stdexcept>
#include <vector>
#include <cmath>
#include <cstdlib>

#ifdef FINDER_GET_EXECUTION_TIME
    #include <stopwatch.h>
#endif

using namespace finder;

#ifndef FINDER_LIGHT_APIX
const int finder::apix::nothing          = 0x00000000;
const int finder::apix::scan_high2low    = 0x00000001;
const int finder::apix::scan_low2high    = 0x00000002;
const int finder::apix::possive_center   = 0x00000004;
const int finder::apix::stripe_testp     = 0x00000008;
const int finder::apix::droped           = 0x00000010;
const int finder::apix::asstripe         = 0x00000020;
const int finder::apix::bycircletest     = 0x00000040;
const int finder::apix::distance_line    = 0x00000080;
#endif

//定数
const int finder::light_threshold = 60;
const int finder::stage1_allowed_distfluc = 3;
const double finder::stage1_allowed_distfluc_rate = 0.2;

const int finder::stage2_ignoreable = 2;
const double finder::stage2_ignoreable_rate = 0.1;
const int finder::stage2_allowed_distfluc = 3;
const double finder::stage2_allowed_distfluc_rate = 0.3;

const double finder::merge_radius_fluc_rate = 0.3;
const double finder::merge_distance_rate = 0.3;

const int finder::circletest_ignoreable = 5;
const double finder::circletest_ignoreable_rate = 0.1;
const int finder::circletest_allowed_distfluc = 3;
const double finder::circletest_allowed_distfluc_rate = 0.3;

const double finder::direction_accuracy = M_PI / 512;
const int finder::direction_marker_maxnum = 4;
const int finder::direction_marker_w = 28;
const int finder::direction_marker_w_dmax = 10;
const int finder::direction_marker_w_min
                            = direction_marker_w - direction_marker_w_dmax;
const int finder::direction_marker_w_max
                            = direction_marker_w + direction_marker_w_dmax;

const int snap_transition_width = 2;
const int snap_allowed_cross_distortion = 2;
const double snap_allowed_cross_distortion_rate = 1;
const int snap_allowed_linear_distortion = 2;
const double snap_allowed_linear_distortion_rate = 0.4;

finder::axistrans::~axistrans(){
}

point2d finder::axistrans::transform(point2d &origin){
    marker orig;

    orig.x = origin.x;
    orig.y = origin.y;

    return transform(orig);
}

finder::simple_axistrans::simple_axistrans
(point2d from1, point2d from2, point2d to1, point2d to2){
    x_rate = (to1.x-to2.x)/(from1.x-from2.x);
    y_rate = (to1.y-to2.y)/(from1.y-from2.y);
    x_diff = (from1.x*to1.x-from1.x*to2.x)/(from1.x-from2.x);
    y_diff = (from1.y*to1.y-from1.y*to2.y)/(from1.y-from2.y);
    xy_rate = y_rate / x_rate;
    rate = (x_rate + y_rate) / 2;
}

finder::simple_axistrans::~simple_axistrans(){
}

point2d finder::simple_axistrans::transform(point2d &origin){
    finder::point2d ans;

    ans.x = origin.x * x_rate + x_diff;
    ans.y = origin.y * y_rate + y_diff;

    return ans;
}

marker finder::simple_axistrans::transform(marker &origin){
    marker ans;

    ans.x = origin.x * x_rate + x_diff;
    ans.y = origin.y * y_rate + y_diff;
    ans.r = origin.r * rate;
    ans.direction = std::atan(std::tan(origin.direction) * xy_rate);
    if(origin.direction > M_PI/2){
        ans.direction += M_PI;
    }
    if(ans.direction < 0){
        ans.direction += 2*M_PI;
    }
    ans.gray = origin.gray;;
    ans.gray_int = origin.gray_int;
    ans.id = origin.id;

    return ans;
}

finder::no_axistrans::no_axistrans(){
}

finder::no_axistrans::~no_axistrans(){
}

point2d finder::no_axistrans::transform(point2d &origin){
    return origin;
}

marker finder::no_axistrans::transform(marker &origin){
    return origin;
}

finder::finder_context::finder_context(int _w, int _h)
: w(_w), h(_h){
    bitmap = new apix[_w * _h];
    have_atrans = true;
    atrans = new no_axistrans();
}

finder::finder_context::~finder_context(){
    delete [] bitmap;
    if(have_atrans){
        delete atrans;
    }
}

void finder::finder_context::set_frame(const apix *fdat){
    while(markers_prev.size() >= 10){
        markers_prev.erase(markers_prev.begin());
    }
    markers_prev.push_back(markers);
    markers.clear();

    for(int i = 0; i < w * h; ++i){
        bitmap[i] = fdat[i];
    }
}

void finder::finder_context::set_frame(const std::string &rgb_string){
    if(rgb_string.size() != w*h*3){
        throw std::out_of_range("size of rgb_string");
    }

    while(markers_prev.size() >= 10){
        markers_prev.erase(markers_prev.begin());
    }
    markers_prev.push_back(markers);
    markers.clear();

    std::string::const_iterator p = rgb_string.begin();
    for(int i = 0; i < w*h; ++i){
#ifdef FINDER_LIGHT_APIX
        uint8_t r = *(p++);
        uint8_t g = *(p++);
        uint8_t b = *(p++);
        bitmap[i].gscale = r + g + b;
#else
        bitmap[i].color.r = *(p++);
        bitmap[i].color.g = *(p++);
        bitmap[i].color.b = *(p++);
        bitmap[i].gscale
                    = bitmap[i].color.r + bitmap[i].color.g + bitmap[i].color.b;
        bitmap[i].flags = 0;
#endif
    }
}

void finder::finder_context::track_all(){
#ifdef FINDER_GET_EXECUTION_TIME
    stopwatch::watch wins;
    wins.plot();
#endif
    mark_possibility(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif

    //直前のマーカーの位置から、追加でマーカー候補点を大量投入
    add_tracking_point(markers, markers_prev);
    //もっと雑なgrayのgetを追加すべき
    get_marker_gray(bitmap, w, h, markers);
    snap_markers(bitmap, w, h, markers);

#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    drop_stripe(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    merge_markers(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    get_marker_gray(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    snap_markers(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    marker_test_circle(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    decide_direction(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
    wins.print();
#endif
}

void finder::finder_context::find_all(){
#ifdef FINDER_GET_EXECUTION_TIME
    stopwatch::watch wins;
    wins.plot();
#endif
    mark_possibility(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    drop_stripe(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    merge_markers(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    get_marker_gray(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    snap_markers(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    marker_test_circle(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
#endif
    decide_direction(bitmap, w, h, markers);
#ifdef FINDER_GET_EXECUTION_TIME
    wins.plot();
    wins.print();
#endif
}

void finder::finder_context::set_axistrans(axistrans &context){
    if(have_atrans){
        delete atrans;
    }
    have_atrans = false;
    atrans = &context;
}

void finder::finder_context::set_axistrans(axistrans *context){
    if(have_atrans){
        delete atrans;
    }
    have_atrans = true;
    atrans = context;
}

std::vector<marker> finder::finder_context::get_markers(){
    std::vector<marker> ret;
    get_markers(ret);
    return ret;
}

void finder::finder_context::get_markers(std::vector<marker> &ret){
    std::vector<marker>::iterator p = markers.begin();
    std::vector<marker>::iterator end = markers.end();
    for(p; p != end; ++p){
        ret.push_back(atrans->transform(*p));
    }
}

size_t finder::finder_context::get_marker_num(){
    return markers.size();
}

//マーカーの位置を、ある程度絞る。
//左→右に走査して、明るさがHLHLHLHLHになっている場所を探す。
//実際は明るさの変化する点を見つけてる
//要修正?: マーカー内部にノイズがあるとマーカーを見つけられないことがある
void finder::mark_possibility
(apix *fdat, int w, int h, std::vector<marker> &markers){
    for(int y = 0; y < h; ++y){
        apix *home = &fdat[y * w];
        apix *pixel = home;
        int last_gray = pixel->gscale;
        int ren_detect = 0;

        std::vector<int> marked;
        std::vector<int> distance;
        std::vector<int> dist2markindex;
        int last_c = 0;

        //一列分の、明暗の変化点をマークする
        for(int x = 1; x < w; ++x){
            ++pixel;

            if(last_gray < pixel->gscale - light_threshold){
                if(ren_detect < 1){
#ifndef FINDER_LIGHT_APIX
                    pixel->flags |= apix::scan_low2high;
#endif
                    ren_detect = 1;

                    if(last_c == -1){
                        distance.push_back(x - marked.back());
                        dist2markindex.push_back(marked.size());
                    }
                    marked.push_back(x);

                    last_c = 1;
                }
            }else if(last_gray > pixel->gscale + light_threshold){
                if(ren_detect > -1){
#ifndef FINDER_LIGHT_APIX
                    pixel->flags |= apix::scan_high2low;
#endif
                    ren_detect = -1;

                    if(last_c == 1){
                        distance.push_back(x - marked.back());
                        dist2markindex.push_back(marked.size());
                    }
                    marked.push_back(x);

                    last_c = -1;
                }
            }else{
                ren_detect = 0;
            }

            //今回使用した値:このピクセルの明るさ=1:2 の加重平均(雑)
            last_gray /= 3;
            last_gray += pixel->gscale * 2 / 3;
        }

        //マークした明暗の変化点から、それらしいものを探す
        //イテレータ使わないほうが良かった気がするから、そのうち書き換えるかも
        if(distance.size() >= 7){
            std::vector<int>::iterator begin = distance.begin();
            std::vector<int>::iterator end = distance.end();
            std::vector<int>::iterator p1 = begin;
            std::vector<int>::iterator stop1 = end - 6;
            for(; p1 != stop1; ++p1){
                int sum = 0;
                std::vector<int>::iterator p2 = p1;
                std::vector<int>::iterator stop2 = p1 + 7;
                for(; p2 != stop2; ++p2){
                    sum += *p2;
                }
                double average = sum / 7.0;
                int maxdist = average*(1 + stage1_allowed_distfluc_rate) + 0.5;
                int mindist = average*(1 - stage1_allowed_distfluc_rate) + 0.5;
                if(maxdist < average + stage1_allowed_distfluc){
                    int maxdist = average + stage1_allowed_distfluc + 0.5;
                }
                if(mindist > average - stage1_allowed_distfluc){
                    int mindist = average - stage1_allowed_distfluc + 0.5;
                }

                p2 = p1;
                stop2 = p1 + 7;
                for(; p2 != stop2; ++p2){
                    if(!(mindist <= *p2 && *p2 <= maxdist)){
                        break;
                    }
                }
                if(p2 == stop2){
                    marker m;
                    m.r = marked.at(dist2markindex.at(p1 - begin + 6))
                          - marked.at(dist2markindex.at(p1 - begin) - 1);
                    m.r /= 2;
                    m.y = y;
                    m.x = marked.at(dist2markindex.at(p1 - begin) - 1)
                          + m.r;
#ifndef FINDER_LIGHT_APIX
                    home[int(m.x+0.5)].flags |= apix::possive_center;
#endif
                    markers.push_back(m);
                }
            }
        }
    }
}

//縦方向の明暗を確認することでマーカー候補を更に絞る。
//格子模様は除外できない。
//円の線の太さを越える大きさの余白が外周に必要です。
void
finder::drop_stripe(apix *fdat, int w, int h, std::vector<marker> &markers){
    std::vector<marker>::iterator p = markers.begin();
    for(p; p != markers.end(); ++p){
        int x;
        int start, stop;
        int gray;
        int c1, c2;

        //円の線の太さ。
        //細すぎる奴は(候補に入っている可能性無い気がするけど)はねておく。
        double thickness = p->r / 3.5;
        if(thickness <= 1){
            goto drop;
        }

        //余白を確認(足りなければdrop)
        if(p->x - thickness * 4.5 < 1 || p->x + thickness * 4.5 > w - 2
        || p->y - thickness * 4.5 < 1 || p->y + thickness * 4.5 > h - 2){
            goto drop;
        }

        //縦方向各ピクセルの、明るさの平均を求めておく。
        //黒と白のちょうど中央の値に近づけるために、
        //一番下の黒帯は含めない
        x = p->x + 0.5;
        start = p->y - p->r + 0.5;
        stop = p->y + thickness * 2.5;
        gray = 0;
        for(int y = start; y <= stop; ++y){
            gray += fdat[y*w + x].gscale;
        }
        gray /= stop - start + 1;

        //各線の中心点の色を確認
        c1 = 0;
        for(int i = -4; i <= 4; ++i){
            int y = p->y + thickness * i;
            if(i & 1){
                if(fdat[y*w + x].gscale > gray){
                    ++c1;
                    if(stage2_ignoreable < c1){
                        goto drop;
                    }
                }
            }else{
                if(fdat[y*w + x].gscale <= gray){
                    ++c1;
                    if(stage2_ignoreable < c1){
                        goto drop;
                    }
                }
            }
        }

        start = p->y - thickness * 4.5 + 0.5;
        stop = p->y + thickness * 4.5 + 0.5;
        c1 = 0;
        c2 = 0;
        //中心以外も確認
        for(int d = 0; d < stop - start; ++d){
            int y = start + d;
            int lv = d / thickness;
            double z = d / thickness - lv;
            if(    z < stage2_allowed_distfluc_rate
                || z * thickness < stage2_allowed_distfluc
                || z > 1 - stage2_allowed_distfluc_rate
                || (1-z) * thickness < stage2_allowed_distfluc){
                continue;
            }
#ifndef FINDER_LIGHT_APIX
            fdat[y*w + x].flags |= apix::stripe_testp;
#endif
            ++c1;
            if(lv & 1){
                if(fdat[y*w + x].gscale > gray){
                    ++c2;
                }
            }else{
                if(fdat[y*w + x].gscale <= gray){
                    ++c2;
                }
            }
        }
        if(c1 && stage2_ignoreable < c2
              && stage2_ignoreable_rate < double(c2) / c1){
            goto drop;
        }

        continue;

drop:
#ifndef FINDER_LIGHT_APIX
        fdat[int(p->y+0.5)*w + int(p->x+0.5)].flags |= apix::droped;
        fdat[int(p->y+0.5)*w + int(p->x+0.5)].flags |= apix::asstripe;
#endif
        p = markers.erase(p) - 1;
    }
}

//近くにある中心候補点をまとめる
void
finder::merge_markers(apix *fdat, int w, int h, std::vector<marker> &markers){
    std::vector<marker> centers;
    std::vector< std::vector<int> > groups;

    int original_size = markers.size();
    for(int i = 0; i < original_size; ++i){
        int group_number = groups.size();
        int j;
        for(j = 0; j < group_number; ++j){
            double rsub = centers[j].r - markers[i].r;
            if(rsub < 0){
                rsub = -rsub;
            }
            if(rsub > centers[j].r * merge_radius_fluc_rate){
                continue;
            }

            double xdist = centers[j].x - markers[i].x;
            double ydist = centers[j].y - markers[i].y;
            double dist2 = xdist * xdist + ydist * ydist;
            if(dist2 > centers[j].r * merge_distance_rate *
                       centers[j].r * merge_distance_rate){
                continue;
            }

            groups[j].push_back(i);
            double xsum = 0;
            double ysum = 0;
            double rsum = 0;
            std::vector<int>::iterator p = groups[j].begin();
            std::vector<int>::iterator end = groups[j].end();
            for(p; p != end; ++p){
                xsum += markers[*p].x;
                ysum += markers[*p].y;
                rsum += markers[*p].r;
            }
            int num = groups[j].size();
            centers[j].x = xsum / num;
            centers[j].y = ysum / num;
            centers[j].r = rsum / num;
            break;
        }
        if(j == group_number){
            centers.push_back(markers[i]);
            std::vector<int> newgroup;
            newgroup.push_back(i);
            groups.push_back(newgroup);
        }
    }
    markers = centers;
}

//マーカーの中心点をぴったりにする
void
finder::snap_markers(apix *fdat, int w, int h, std::vector<marker> &markers){
    std::vector<marker>::iterator p = markers.begin();
    for(p; p != markers.end(); ++p){
        double thickness = p->r / 3.5;
        //int left, right, top, bottom;
        //int dx, dy;
        int lx, ly;
        int pos[4];
        struct{int x,y;} dir[4] = {{1,0}, {-1,0}, {0,1}, {0,-1}};
        int i;

        if(p->x - thickness < 0 || p->x + thickness >= w ||
           p->y - thickness < 0 || p->y + thickness >= h){
            goto drop;
        }

        if(fdat[int(p->y+0.5)*w + int(p->x+0.5)].gscale <= p->gray_int){
            //中心点が黒かったら、ちょっとずれた位置で試す
            int i;
            struct{int x,y;} ptt[] = {{1, 0}, {0, 1}, {-1, 0}, {0, -1},
                                     {1, 1}, {1, -1}, {-1, 1}, {-1, -1}};
            int delta = thickness / 20;
            if(delta <= 0 && thickness / 10 > 1){
                delta = 1;
            }

            for(i = 0; i < sizeof(ptt)/sizeof(ptt[0]); ++i){
                int dx = delta * ptt[i].x;
                int dy = delta * ptt[i].y;
                int x = p->x + dx;
                int y = p->y + dy;

                if(fdat[y*w + x].gscale > p->gray_int){
                    p->x = x;
                    p->y = y;
                    break;
                }
            }
            if(i == sizeof(ptt)/sizeof(ptt[0])){
                goto drop;
            }
        }

//よく考えたら、ここらへんいらない気がする。ということで、とりあえず無効化
#if 0
        //中心部の白い部分を使って、大まかにsnap
        //左側を見つける
        dx = 0;
        while(true){
            --dx;
            int x = p->x + dx + 0.5;
            if(dx <= -thickness * 2.5 || x < 0){
                goto drop;
            }
            if(fdat[int(p->y+0.5)*w + x].gscale <= p->gray_int){
                left = x;
                break;
            }
        }
        //右側を見つける
        dx = 0;
        while(true){
            ++dx;
            int x = p->x + dx + 0.5;
            if(dx >= thickness * 2.5 || x >= w){
                goto drop;
            }
            if(fdat[int(p->y+0.5)*w + x].gscale <= p->gray_int){
                right = x;
                break;
            }
        }
        p->x = (left + right) / 2.0;
        //上側を見つける
        dy = 0;
        while(true){
            --dy;
            int y = p->y + dy + 0.5;
            if(dy <= -thickness * 2.5 || y < 0){
                goto drop;
            }
            if(fdat[y*w + int(p->x+0.5)].gscale <= p->gray_int){
                top = y;
                break;
            }
        }
        //下側を見つける
        dy = 0;
        while(true){
            ++dy;
            int y = p->y + dy + 0.5;
            if(dy >= thickness * 2.5 || y >= h){
                goto drop;
            }
            if(fdat[y*w + int(p->x+0.5)].gscale <= p->gray_int){
                bottom = y;
                break;
            }
        }
        p->y = (top + bottom) / 2.0;
#endif

        //最外周を使って中心をしっかり合わせる
        for(i = 0; i < 4; ++i){
            int last_color = 1;
            int ren = 0;
            int lines = 0;
            int delta;
            for(delta = 0; delta < thickness * 5; ++delta){
                int x = p->x + delta * dir[i].x;
                int y = p->y + delta * dir[i].y;

                if(x < 0 || x >= w || y < 0 || y >= h){
                    goto drop;
                }

                if((fdat[y*w + x].gscale < p->gray_int) - last_color){
                    //色の変化なし
                    ren = 0;
                }else{
                    //色が変化
                    if(++ren >= snap_transition_width){
                        if(++lines == 4){
                            pos[i] = delta - ren + 1;
                            break;
                        }
                        last_color = !last_color;
                        ren = 0;
                    }
                }
            }
            if(delta >= thickness * 5){
                goto drop;
            }
        }

        //極端に歪んだマーカは排除
        lx = pos[0] + pos[1];
        ly = pos[2] + pos[3];
        if(snap_allowed_cross_distortion < std::abs(lx - ly) &&
           snap_allowed_cross_distortion_rate < std::abs((lx-ly)/(lx+ly))*2){
            goto drop;
        }
        if(snap_allowed_linear_distortion < std::abs(pos[0]-pos[1]) &&
           snap_allowed_linear_distortion_rate
                    < std::abs((pos[0]-pos[1])/(pos[0]+pos[1]))*2){
            goto drop;
        }
        if(snap_allowed_linear_distortion < std::abs(pos[2]-pos[3]) &&
           snap_allowed_linear_distortion_rate
                    < std::abs((pos[2]-pos[3])/(pos[2]+pos[3]))*2){
            goto drop;
        }

        p->x += (pos[0] - pos[1]) / 2.0;
        p->y += (pos[2] - pos[3]) / 2.0;
        p->r = (pos[0] + pos[1] + pos[2] + pos[3] - 2) / 4.0;
        continue;

drop:
        p = markers.erase(p) - 1;
    }
}

//マーカーの、grayのレベルを確定させる
void
finder::get_marker_gray(apix *fdat, int w, int h, std::vector<marker> &markers){
    std::vector<marker>::iterator p = markers.begin();
    for(p; p != markers.end(); ++p){
        //円の線の太さ。
        double thickness = p->r / 3.5;

        int boxsize = thickness * 8 + 1;
        int boxx    = p->x - thickness * 4;
        int boxy    = p->y - thickness * 4;

        //余白を確認(足りなければdrop)
        if(boxx < 0 || boxx + boxsize >= w || boxy < 1 || boxy + boxsize >= h){
            p = markers.erase(p) - 1;
            continue;
        }

        //中心からthickness*4以内の範囲の平均がgrayのはず。
        double r2 = thickness * 4 * thickness * 4;
        long long int sum = 0;
        int points = 0;
        for(int y = boxy; y < boxy + boxsize; ++y){
            for(int x = boxx; x < boxx + boxsize; ++x){
                double dx = x - p->x;
                double dy = y - p->y;
                double d2 = dx * dx + dy * dy;
                if(d2 < r2){
                    ++points;
                    sum += fdat[y*w + x].gscale;
                }
            }
        }
        p->gray = double(sum) / points;
        p->gray_int = static_cast<int>(p->gray);
    }
}

void finder::marker_test_circle
(apix *fdat, int w, int h, std::vector<marker> &markers){
    std::vector<marker>::iterator p = markers.begin();
    for(p; p != markers.end(); ++p){
        int ystart, ystop;
        int c1, c2;

        //円の線の太さ。
        //細すぎる奴は(候補に入っている可能性無い気がするけど)はねておく。
        double thickness = p->r / 3.5;
        if(thickness <= 1){
            goto drop;
        }

        //余白を確認(足りなければdrop)
        if(p->x - thickness * 4.5 < 1 || p->x + thickness * 4.5 > w - 2
        || p->y - thickness * 4.5 < 1 || p->y + thickness * 4.5 > h - 2){
            goto drop;
        }

        c1 = c2 = 0;
        ystart = p->y - thickness * 4.5;
        ystop = p->y + thickness * 4.5 + 1;
        for(int y = ystart; y <= ystop; ++y){
            int xstart = p->x - thickness * 4.5;
            int xstop = p->x + thickness * 4.5 + 1;
            for(int x = xstart; x <= xstop; ++x){
                double rpos = (x-p->x)*(x-p->x)+(y-p->y)*(y-p->y);
                rpos = std::sqrt(rpos);
                double d = thickness * 4.5 - rpos;

                if(rpos > thickness * 4.5){
                    continue;
                }

                int lv = d / thickness;
                double z = d / thickness - lv;
                if(    z < circletest_allowed_distfluc_rate
                    || z * thickness < circletest_allowed_distfluc
                    || z > 1 - circletest_allowed_distfluc_rate
                    || (1-z) * thickness < circletest_allowed_distfluc){
                    continue;
                }
                ++c1;
                if(lv & 1){
                    if(fdat[y*w + x].gscale > p->gray_int){
                        ++c2;
                    }
                }else{
                    if(fdat[y*w + x].gscale <= p->gray_int){
                        ++c2;
                    }
                }
            }
        }
        if(c1 && circletest_ignoreable < c2
              && circletest_ignoreable_rate < double(c2) / c1){
            goto drop;
        }

        continue;

drop:
#ifndef FINDER_LIGHT_APIX
        fdat[int(p->y+0.5)*w + int(p->x+0.5)].flags |= apix::droped;
        fdat[int(p->y+0.5)*w + int(p->x+0.5)].flags |= apix::bycircletest;
#endif
        p = markers.erase(p) - 1;
    }
}

//方向とIDを特定する
//だいぶクズなコードですね。わけわかんねぇ
void finder::decide_direction
(apix *fdat, int w, int h, std::vector<marker> &markers){
    std::vector<marker>::iterator p = markers.begin();
    for(p; p != markers.end(); ++p){
        double mark_d = p->r / 3.5 * 5;
        p->id = 0;

        //余白を確認
        if(p->x - mark_d < 1 || p->x + mark_d > w - 2
        || p->y - mark_d < 1 || p->y + mark_d > h - 2){
            continue;
        }

        int last_x = -1;
        int last_y = -1;

        int last_c = 0;         //0=B, 1=W
        int c_stab = -1;        //0=B, 1=W

        int last_c_stab = -1;   //0=B, 1=W
        int color_d = 0;
        int conn_ms = 0;
        double conn_start_d = 0;
        double conn_tail_d = 0;
        int cdstack = 0;
        int best_dvalue = 5000;
        int dvalue = 0;
        for(double d = 0; d < M_PI * 4; d += direction_accuracy){
            int x = p->x + mark_d * std::cos(d) + 0.5;
            int y = p->y + mark_d * std::sin(d) + 0.5;
            ++cdstack;
            if(last_x == x && last_y == y){
                continue;
            }
            last_x = x;
            last_y = y;
#ifndef FINDER_LIGHT_APIX
            fdat[y*w + x].flags |= apix::distance_line;
#endif
            int c = fdat[y*w + x].gscale > p->gray_int;
            if(last_c == c | c_stab < 0){
                c_stab = c;
            }
            last_c = c;

            if(last_c_stab == c_stab){
                color_d += cdstack;
                if(color_d > direction_marker_w_max && c_stab && conn_ms & 1
                        && (conn_ms+1)/2 <= direction_marker_maxnum
                        && conn_start_d > direction_accuracy
                             *direction_marker_maxnum*direction_marker_w_max*2){
                    if(dvalue < best_dvalue){
                        p->id = (conn_ms + 1) / 2;
                        p->direction
                            = conn_start_d + (conn_tail_d-conn_start_d) / 2;
                        while(p->direction >= M_PI * 2){
                            p->direction -= M_PI * 2;
                        }
                        p->direction -= direction_accuracy;
                        best_dvalue = dvalue;
                    }
                }
            }else{
#ifndef FINDER_LIGHT_APIX
                if(c_stab){
                    fdat[y*w + x].flags |= apix::scan_low2high;
                }else{
                    fdat[y*w + x].flags |= apix::scan_high2low;
                }
#endif
                if(direction_marker_w_min <= color_d
                        && color_d <= direction_marker_w_max){
                    ++conn_ms;
                    conn_tail_d = d;
                    int don = direction_marker_w - color_d;
                    if(don < 0){
                        don = -don;
                    }
                    if(dvalue < don){
                        dvalue = don;
                    }
                }else{
                    dvalue = 0;
                    conn_ms = 0;
                    conn_start_d = d;
                }
                color_d = 0;
            }
            last_c_stab = c_stab;
            cdstack = 0;
        }
    }
}

//markers_prevのマーカーの位置を少しずらしたような物をmarkersについか
void finder::add_tracking_point
(std::vector<marker> &markers,
 const std::vector< std::vector<marker> > &markers_prev){
    std::vector< std::vector<marker> >::const_iterator pp
            = markers_prev.begin();
    std::vector< std::vector<marker> >::const_iterator endp
            = markers_prev.end();
    for(; pp < endp; ++pp){
        std::vector<marker>::const_iterator p = pp->begin();
        std::vector<marker>::const_iterator end = pp->end();
        for(; p < end; ++p){
            for(int dy = -5; dy <= 5; ++dy){
                for(int dx = -5; dx <= 5; ++dx){
                    marker nem = *p;
                    nem.x += nem.r * dx / 5;
                    nem.y += nem.r * dy / 5;
                    markers.push_back(nem);
                }
            }
        }
    }
}
