// Copyright © 2015, 2016 T-semi
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <stdexcept>

#include <getopt.h>             /* getopt_long() */

#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <linux/videodev2.h>

#include <jpeglib.h>

#include "v4l2.h"

#include <iostream>

#define CLEAR(x) ::memset(&(x), 0, sizeof(x))

namespace v4l2{
    static void throw_jpeg_decode_error(j_common_ptr cinfo){
        (*cinfo->err->output_message)(cinfo);
        throw std::runtime_error("Error detected while decoding JPEG.");
    }

    void camera::free_buffer(){
        std::vector<buffer>::iterator p = bufs.begin();
        std::vector<buffer>::iterator end = bufs.end();
        while(p != end){
            if(io == IO_METHOD_MMAP){
                if(munmap(p->start, p->length) == -1){
                    throw std::runtime_error("munmap error.");
                }
            }else{
                free(p->start);
            }
            ++p;
        }
        bufs.clear();
    }

    void camera::initio_readwrite(){
        buffer b;
        b.length = pix_fmt.sizeimage;
        b.start = malloc(b.length);
        if(b.start == NULL){
            throw std::bad_alloc();
        }
        bufs.push_back(b);
    }

    void camera::initio_mmap(){
        v4l2_requestbuffers req;
        CLEAR(req);

        req.count = 4;
        req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        req.memory = V4L2_MEMORY_MMAP;

        if(-1 == ::ioctl(fd, VIDIOC_REQBUFS, &req)){
            if(EINVAL == errno){
                throw std::runtime_error("Device does not support MMAP IO.");
            }else{
                throw std::runtime_error("Unknown VIDIOC_REQBUFS error");
            }
        }

        if(req.count < 2){
            throw std::runtime_error("Insufficient buffer memory.");
        }

        for(int i = 0; i < req.count; ++i){
            v4l2_buffer buf;

            CLEAR(buf);

            buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory      = V4L2_MEMORY_MMAP;
            buf.index       = i;

            if (-1 == ::ioctl(fd, VIDIOC_QUERYBUF, &buf)){
                free_buffer();
                throw std::runtime_error("VIDIOC_QUERYBUF");
            }

            buffer b;
            b.length = buf.length;
            b.start = mmap(NULL /* start anywhere */,
                           b.length,
                           PROT_READ | PROT_WRITE /* required */,
                           MAP_SHARED /* recommended */,
                           fd, buf.m.offset);
            if(MAP_FAILED == b.start){
                free_buffer();
                throw std::bad_alloc();
            }
            bufs.push_back(b);
        }

        for(int i = 0; i < req.count; ++i){
            v4l2_buffer buf;

            CLEAR(buf);
            buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory = V4L2_MEMORY_MMAP;
            buf.index = i;

            if (-1 == ::ioctl(fd, VIDIOC_QBUF, &buf)){
                free_buffer();
                throw std::runtime_error("VIDIOC_QBUF");
            }
        }
        v4l2_buf_type type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        if(-1 == ::ioctl(fd, VIDIOC_STREAMON, &type)){
            free_buffer();
            throw std::runtime_error("VIDIOC_STREAMON");
        }
    }

    void camera::initio_userptr(){
        v4l2_requestbuffers req;

        CLEAR(req);

        req.count  = 4;
        req.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        req.memory = V4L2_MEMORY_USERPTR;

        if(-1 == ::ioctl(fd, VIDIOC_REQBUFS, &req)){
            if(EINVAL == errno){
                throw std::runtime_error("Device does not support userptr IO.");
            }else{
                throw std::runtime_error("Unknown VIDIOC_REQBUFS error");
            }
        }

        for(int i = 0; i < 4; ++i){
            buffer b;
            b.length = pix_fmt.sizeimage;
            b.start = malloc(b.length);

            if(!b.start){
                free_buffer();
                throw std::bad_alloc();
            }

            bufs.push_back(b);
        }

        int i = 0;
        std::vector<buffer>::iterator p = bufs.begin();
        std::vector<buffer>::iterator end = bufs.end();
        while(p != end){
            v4l2_buffer buf;

            CLEAR(buf);
            buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory = V4L2_MEMORY_USERPTR;
            buf.index = i;
            buf.m.userptr = (unsigned long)(p->start);
            buf.length = p->length;

            if(-1 == ::ioctl(fd, VIDIOC_QBUF, &buf)){
                free_buffer();
                throw std::runtime_error("VIDIOC_QBUF error");
            }
            ++p;
            ++i;
        }
        v4l2_buf_type type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        if(-1 == ::ioctl(fd, VIDIOC_STREAMON, &type)){
            free_buffer();
            throw std::runtime_error("VIDIOC_STREAMON error");
        }
    }

    camera::camera(std::string _dev_name):
            fd(-1),
            initialized(false),
            io(IO_METHOD_MMAP),
            dev_name(_dev_name){

        CLEAR(pix_fmt);

        pix_fmt.width       = 640;
        pix_fmt.height      = 480;
        pix_fmt.pixelformat = V4L2_PIX_FMT_YUYV;
        pix_fmt.field       = V4L2_FIELD_NONE;

        CLEAR(parm);
    }

    camera::~camera(){
        if(initialized){
            uninit();
        }
        if(fd != -1){
            close();
        }
    }

    void camera::open(){
        struct stat st;

        if(fd != -1){
            throw std::runtime_error("Already opened");
        }

        if(-1 == ::stat(dev_name.c_str(), &st)){
            throw std::runtime_error("Cannot identify the file");
        }

        if(!S_ISCHR(st.st_mode)){
            throw std::runtime_error("Not a device file");
        }

        fd = ::open(dev_name.c_str(), O_RDWR /* required */ | O_NONBLOCK, 0);

        if(-1 == fd){
            throw std::runtime_error("Opening device failed");
        }

        if(-1 == ::ioctl(fd, VIDIOC_QUERYCAP, &cap)){
            if(EINVAL == errno){
                throw std::runtime_error("Not a video device file");
            }else{
                throw std::runtime_error("VIDIOC_QUERYCAP error");
            }
        }
        if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
            throw std::runtime_error("Not a video capture device");
        }

        CLEAR(cropcap);
        cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        if(0 == ::ioctl(fd, VIDIOC_CROPCAP, &cropcap)){
            crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            crop.c = cropcap.defrect; /* reset to default */

            if(-1 == ::ioctl(fd, VIDIOC_S_CROP, &crop)){
                if(errno == EINVAL){
                    /* Cropping not supported. */
                }else{
                    /* Errors ignored. */
                }
            }
        }else{
            /* Errors ignored. */
        }
    }

    void camera::close(){
        if(fd == -1){
            throw std::runtime_error("Already closed");
        }
        if(::close(fd) == -1){
            throw std::runtime_error("Closing device failed");
        }
        fd = -1;
    };

    void camera::init(){
        if(fd == -1){
            throw std::runtime_error("devicefile not opened");
        }else if(initialized){
            throw std::runtime_error("already initialized");
        }

        if(io == IO_METHOD_READ){
            if (!(cap.capabilities & V4L2_CAP_READWRITE)) {
                throw std::runtime_error("IO method not supported");
            }
        }else{
            if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
                throw std::runtime_error("IO method not supported");
            }
        }

        v4l2_format fmt;
        fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        fmt.fmt.pix = pix_fmt;

        v4l2_streamparm stp;
        stp.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        stp.parm.capture = parm;

        if (-1 == ::ioctl(fd, VIDIOC_S_FMT, &fmt)){
            if(errno == EINVAL){
                throw std::runtime_error("Feature not supported");
            }else{
                throw std::runtime_error("Unknown VIDIOC_S_FMT error.");
            }
        }

        if (-1 == ::ioctl(fd, VIDIOC_S_PARM, &stp)){
            if(errno == EINVAL){
                throw std::runtime_error("Feature not supported");
            }else{
                throw std::runtime_error("Unknown VIDIOC_S_PARM error.");
            }
        }
        parm = stp.parm.capture;

        fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        if (-1 == ::ioctl(fd, VIDIOC_G_FMT, &fmt)){
            throw std::runtime_error("Unknown VIDIOC_G_FMT error.");
        }
        pix_fmt = fmt.fmt.pix;

        int min;
        min = pix_fmt.width * 2;
        if (pix_fmt.bytesperline < min){
            pix_fmt.bytesperline = min;
        }
        min = pix_fmt.bytesperline * pix_fmt.height;
        if (pix_fmt.sizeimage < min){
            pix_fmt.sizeimage = min;
        }

        if(io == IO_METHOD_READ){
            initio_readwrite();
        }else if(io == IO_METHOD_MMAP){
            initio_mmap();
        }else if(io == IO_METHOD_USERPTR){
            initio_userptr();
        }

        initialized = true;
    }

    void camera::uninit(){
        if(!initialized){
            throw std::runtime_error("not initialized.");
        }

        if(io != IO_METHOD_READ){
            v4l2_buf_type type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            if(-1 == ::ioctl(fd, VIDIOC_STREAMOFF, &type)){
                throw std::runtime_error("VIDIOC_STREAMOFF error");
            }
        }
        free_buffer();
        initialized = false;
    }

    v4l2_fmtdesc camera::get_format(uint32_t index){
        if(fd == -1){
            throw std::runtime_error("Device closed");
        }

        v4l2_fmtdesc fmt;
        fmt.index = index;
        fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        if(::ioctl(fd, VIDIOC_ENUM_FMT, &fmt) == -1){
            if(errno == EINVAL){
                throw std::out_of_range("Format index out of range.");
            }else{
                throw std::runtime_error("VIDIOC_ENUM_FMT");
            }
        }
        return fmt;
    }

    v4l2_frmsizeenum camera::get_frame_size(uint32_t pixel_fmt, uint32_t index){
        if(fd == -1){
            throw std::runtime_error("Device closed");
        }

        v4l2_frmsizeenum fmt;
        fmt.pixel_format = pixel_fmt;
        fmt.index = index;
        if(::ioctl(fd, VIDIOC_ENUM_FRAMESIZES, &fmt) == -1){
            if(errno == EINVAL){
                throw std::invalid_argument("Frame size discriptor invalid.");
            }else{
                throw std::runtime_error("VIDIOC_ENUM_FRAMESIZES");
            }
        }
        return fmt;
    }

    v4l2_frmivalenum camera::get_frame_interval(
            uint32_t pixel_fmt, uint32_t w, uint32_t h, uint32_t index){
        if(fd == -1){
            throw std::runtime_error("Device closed");
        }

        v4l2_frmivalenum fmt;
        fmt.pixel_format = pixel_fmt;
        fmt.width = w;
        fmt.height = h;
        fmt.index = index;
        if(::ioctl(fd, VIDIOC_ENUM_FRAMEINTERVALS, &fmt) == -1){
            if(errno == EINVAL){
                throw std::invalid_argument("Interval discriptor invalid.");
            }else{
                throw std::runtime_error("VIDIOC_ENUM_FRAMEINTERVALS");
            }
        }
        return fmt;
    }

    void camera::set_io_method(io_method _io){
        if(initialized){
            throw std::runtime_error("Operation not allowed for "
                                     "initialized camera");
        }
        io = _io;
    }

    void camera::set_framesize(int w, int h){
        if(initialized){
            throw std::runtime_error("Operation not allowed for "
                                     "initialized camera");
        }
        pix_fmt.width = w;
        pix_fmt.height = h;
    }

    void camera::set_format(uint32_t formatcode){
        if(initialized){
            throw std::runtime_error("Operation not allowed for "
                                     "initialized camera");
        }
        pix_fmt.pixelformat = formatcode;
    }

    void camera::set_frame_interval(v4l2_fract interval){
        if(initialized){
            throw std::runtime_error("Operation not allowed for "
                                     "initialized camera");
        }
        parm.timeperframe = interval;
    }

    void camera::set_frame_interval(uint32_t numerator, uint32_t denominator){
        v4l2_fract interval;
        interval.numerator = numerator;
        interval.denominator = denominator;
        set_frame_interval(interval);
    }

    uint32_t camera::get_format(){
        return pix_fmt.pixelformat;
    }

    v4l2_frmsize_discrete camera::get_frame_size(){
        v4l2_frmsize_discrete val;
        val.width = pix_fmt.width;
        val.height = pix_fmt.height;
        return val;
    }

    v4l2_fract camera::get_frame_interval(){
        return parm.timeperframe;
    }

    size_t camera::get_rawframe_string(std::string &data){
        return get_rawframe_basic_string(data);
    }

    std::string camera::get_rawframe_string(){
        std::string data;
        camera::get_rawframe_string(data);
        return data;
    }

    size_t camera::get_rgbframe_string(std::string &data){
        uint32_t rawframe_format = get_format();
        v4l2_frmsize_discrete fsize = get_frame_size();

        std::string raw;
        if(get_rawframe_string(raw) == 0){
            data.resize(0);
            return 0;
        }

        if(rawframe_format == V4L2_PIX_FMT_YUYV){
            std::string::iterator rp = raw.begin();
            std::string::iterator rend = raw.end();

            data.resize(fsize.width * fsize.height * 3);
            std::string::iterator p = data.begin();
            std::string::iterator end = data.end();

            if((rend - rp) / 2 != (end - p) / 3 || fsize.width % 2){
                std::cout << "fs:" << fsize.width << "," << fsize.height
                          << " bs:" << raw.size() << std::endl;
                throw std::runtime_error("Raw frame broken.");
            }

            while(p != end){
                uint8_t y1 = *(rp++);
                uint8_t u = *(rp++);
                uint8_t y2 = *(rp++);
                uint8_t v = *(rp++);

                uint8_t r, g, b;
                yuv2rgb(y1, u, v, r, g, b);
                *(p++) = r;
                *(p++) = g;
                *(p++) = b;
                yuv2rgb(y2, u, v, r, g, b);
                *(p++) = r;
                *(p++) = g;
                *(p++) = b;
            }
        }else if(rawframe_format == V4L2_PIX_FMT_MJPEG){
            uint8_t *buff_head = const_cast<uint8_t *>(
                                reinterpret_cast<const uint8_t *>(raw.c_str()));
            size_t buff_size = raw.size();

            struct jpeg_error_mgr emgr;
            struct jpeg_decompress_struct cinfo;
            cinfo.err = jpeg_std_error(&emgr);
            emgr.error_exit = throw_jpeg_decode_error;
            data.resize(0);
            data.reserve(fsize.width * fsize.height * 3);
            try{
                jpeg_create_decompress(&cinfo);
                jpeg_mem_src(&cinfo, buff_head, buff_size);
                jpeg_read_header(&cinfo, true);
                jpeg_start_decompress(&cinfo);
                if(cinfo.output_components != 3 ||
                   fsize.width != cinfo.output_width ||
                   fsize.height != cinfo.output_height  ){
                    throw std::runtime_error("Decoder returns invalid value");
                }
                JSAMPARRAY recv_buffer = (*cinfo.mem->alloc_sarray)
                            ((j_common_ptr) &cinfo, JPOOL_IMAGE,
                             cinfo.output_width * cinfo.output_components, 1);
                while(cinfo.output_scanline < cinfo.output_height){
                    jpeg_read_scanlines(&cinfo, recv_buffer, 1);
                    data.append(reinterpret_cast<char *>(recv_buffer[0]),
                                fsize.width * 3);
                }
                jpeg_finish_decompress(&cinfo);
            }catch(std::runtime_error err){
                jpeg_destroy_decompress(&cinfo);
                data = "";
                return 0;
            }
            jpeg_destroy_decompress(&cinfo);
        }else{
            throw std::runtime_error("Not supported conversion.");
        }

        return data.size();
    }

    std::string camera::get_rgbframe_string(){
        std::string data;
        camera::get_rgbframe_string(data);
        return data;
    }

    void camera::wait_frame(){
        ::fd_set fds;
        int ret;

        while(true){
            FD_ZERO(&fds);
            FD_SET(fd, &fds);

            ret = ::select(fd + 1, &fds, NULL, NULL, NULL);
            if(ret == -1){
                if(errno == EINTR){
                    continue;
                }
                throw std::runtime_error("select returns error");
            }else if(ret == 0){
                throw std::runtime_error("select returns invalid value");
            }

            break;
        }
    }

    bool camera::wait_frame(::timeval tv){
        ::fd_set fds;
        int ret;

        while(true){
            FD_ZERO(&fds);
            FD_SET(fd, &fds);

            ret = ::select(fd + 1, &fds, NULL, NULL, &tv);
            if(ret == -1){
                if(errno == EINTR){
                    continue;
                }
                throw std::runtime_error("select returns error");
            }else if(ret == 0){
                //time out
                return false;
            }

            break;
        }
        return true;
    }

    bool camera::wait_frame(long sec, long usec){
        ::timeval tv;
        tv.tv_sec = sec;
        tv.tv_usec = usec;
        return wait_frame(tv);
    }
}
