// Copyright © 2015, 2016 T-semi
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef INCLUDED_V4L2CC_V4L2_H
#define INCLUDED_V4L2CC_V4L2_H

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <stdexcept>

#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <string>
#include <vector>

#include <linux/videodev2.h>

namespace v4l2{
    enum io_method {
        IO_METHOD_READ,
        IO_METHOD_MMAP,
        IO_METHOD_USERPTR,
    };

    struct buffer;

    class wishatom;
    class wisharea;
    template<typename T> class wishrank;
    class wishlist;

    class camera;

    static inline void yuv2rgb(uint8_t y, uint8_t u, uint8_t v,
                               uint8_t &r, uint8_t &g, uint8_t &b){
        int ur = u - 128;
        int vr = v - 128;
        int ro = y            + 1.402*vr + 0.5;
        int go = y - 0.344*ur - 0.714*vr + 0.5;
        int bo = y + 1.772*ur            + 0.5;
        r = ro > 255 ? 255 : (ro < 0 ? 0 : ro);
        g = go > 255 ? 255 : (go < 0 ? 0 : go);
        b = bo > 255 ? 255 : (bo < 0 ? 0 : bo);
    }

    struct buffer{
        void       *start;
        size_t      length;
    };

    class wishatom{
    public:
        std::string label;
        wishatom(std::string label);
        virtual ~wishatom() = 0;
    };

    class wisharea: public wishatom{
    public:
        int best;
        int d;
        wisharea(std::string label, int best, int d);
        ~wisharea();
    };

    template<typename T>
    class wishrank: public wishatom{
    public:
        std::vector<T> list;
        wishrank(std::string label);
        ~wishrank();
        void push(T item);
        T pop();
    };

    class wishlist{
    public:
        std::vector<wishatom *> list;

        wishlist();
        ~wishlist();
    };

    class camera{
    private:
        int                         fd;
        bool                        initialized;
        io_method                   io;
        v4l2_capability             cap;
        v4l2_cropcap                cropcap;
        v4l2_crop                   crop;
        v4l2_pix_format             pix_fmt;
        v4l2_captureparm            parm;
        std::vector<buffer>         bufs;
        void free_buffer();
        void initio_readwrite();
        void initio_mmap();
        void initio_userptr();

        template<typename T>
        class get_rawframe_basic_string_helper{
        private:
            std::basic_string<T> &target;
        public:
            get_rawframe_basic_string_helper(std::basic_string<T> &_target):
                    target(_target){
            }

            void operator ()(buffer b){
                target = std::basic_string<T>(
                        static_cast<T *>(b.start), b.length / sizeof(T));
            }
        };

    public:
        const std::string           dev_name;
        camera(std::string _dev_name);
        ~camera();
        void open();
        void close();
        void init();
        void uninit();
        v4l2_fmtdesc get_format(uint32_t index);
        v4l2_frmsizeenum get_frame_size(uint32_t pixel_fmt, uint32_t index);
        v4l2_frmivalenum get_frame_interval(
                uint32_t pixel_fmt, uint32_t w, uint32_t h, uint32_t index);
        uint32_t get_format();
        v4l2_frmsize_discrete get_frame_size();
        v4l2_fract get_frame_interval();
        void set_io_method(io_method _io);
        void set_format(uint32_t formatcode);
        void set_framesize(int w, int h);
        void set_frame_interval(v4l2_fract interval);
        void set_frame_interval(uint32_t numerator, uint32_t denominator);

        template<typename F>
        size_t get_rawframe_buffer(F frame_handle){
            if(!initialized){
                throw std::runtime_error("not initialized.");
            }

            v4l2_buffer buf;
            buffer b;
            b.length = 0;

            if(io == IO_METHOD_READ){
                b = bufs.at(0);
                if(-1 == ::read(fd, b.start, b.length)){
                    if(errno == EAGAIN){
                        b.length = 0;
                        return 0;
                    }else if(errno != EIO){
                        throw std::runtime_error("read failed.");
                    }
                }
            }else if(io == IO_METHOD_MMAP){
                ::memset(&buf, 0, sizeof(buf));

                buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                buf.memory = V4L2_MEMORY_MMAP;

                if(-1 == ::ioctl(fd, VIDIOC_DQBUF, &buf)){
                    if(errno == EAGAIN){
                        return 0;
                    }else if(errno != EIO){
                        throw std::runtime_error("VIDIOC_DQBUF failed.");
                    }
                }
                b.start = bufs.at(buf.index).start;
                b.length = buf.bytesused;
            }else if(io == IO_METHOD_USERPTR){
                unsigned int i;

                ::memset(&buf, 0, sizeof(buf));

                buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                buf.memory = V4L2_MEMORY_USERPTR;

                if(-1 == ::ioctl(fd, VIDIOC_DQBUF, &buf)){
                    if(errno == EAGAIN){
                        return 0;
                    }else if(errno != EIO){
                        throw std::runtime_error("VIDIOC_DQBUF failed.");
                    }
                }

                std::vector<buffer>::iterator p = bufs.begin();
                std::vector<buffer>::iterator end = bufs.end();
                while(true){
                    if(p == end){
                        throw std::runtime_error(
                                "Device returns uninitialized ptr.");
                    }
                    if(p->start == (void *)(buf.m.userptr)
                            && p->length == buf.length){
                        b = *p;
                        break;
                    }
                    ++p;
                }
            }

            frame_handle(b);

            if(io != IO_METHOD_READ){
                if(-1 == ::ioctl(fd, VIDIOC_QBUF, &buf)){
                    throw std::runtime_error("VIDIOC_QBUF error");
                }
            }

            return b.length;
        }

        template<typename T>
        size_t get_rawframe_basic_string(std::basic_string<T> &data){
            size_t size;
            size = get_rawframe_buffer(
                    get_rawframe_basic_string_helper<T>(data));
            if(!size){
                data.clear();
            }
            return size;
        }

        size_t get_rawframe_string(std::string &data);
        std::string get_rawframe_string();

        size_t get_rgbframe_string(std::string &data);
        std::string get_rgbframe_string();

        void wait_frame();
        bool wait_frame(::timeval tv);
        bool wait_frame(long sec, long usec);
    };
}

#endif  //INCLUDED_V4L2CC_V4L2_H
